import enum
from datetime import date, datetime
from decimal import Decimal

import sqlalchemy as sa
from sqlalchemy import Boolean, Date, DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.types import Enum as SQLEnum


class Base(DeclarativeBase):
    """Base class for all models"""


class UserRole(str, enum.Enum):
    ADMIN = "admin"
    USER = "user"
    GUEST = "guest"


class AccountStatus(int, enum.Enum):
    ACTIVE = 1
    INACTIVE = 0
    PENDING = 2


class Tenant(Base):
    __tablename__ = "tenants"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String)
    accounts: Mapped[list["Account"]] = relationship("Account", back_populates="tenant")


class Actor(Base):
    __tablename__ = "actors"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String)


class Account(Base):
    __tablename__ = "accounts"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String)
    description: Mapped[str | None] = mapped_column(String, nullable=True)
    status: Mapped[AccountStatus] = mapped_column(SQLEnum(AccountStatus))
    balance: Mapped[float] = mapped_column(Float)
    created_at: Mapped[datetime] = mapped_column(DateTime)
    created_by: Mapped[Actor] = relationship(Actor)
    created_by_id: Mapped[int] = mapped_column(ForeignKey("actors.id"))
    tenant_id: Mapped[int] = mapped_column(ForeignKey("tenants.id"), nullable=True)
    tenant: Mapped[Tenant] = relationship("Tenant", back_populates="accounts")
    users: Mapped[list["User"]] = relationship("User", back_populates="account")
    datasource_id: Mapped[str] = mapped_column(String(255), nullable=False)


class User(Base):
    __tablename__ = "users"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String)
    age: Mapped[int] = mapped_column(Integer)
    email: Mapped[str] = mapped_column(String, nullable=True)
    role: Mapped[UserRole] = mapped_column(SQLEnum(UserRole))
    is_active: Mapped[bool] = mapped_column(Boolean)
    birth_date: Mapped[date] = mapped_column(Date)
    account_id: Mapped[int] = mapped_column(ForeignKey("accounts.id"))
    account: Mapped[Account] = relationship("Account", back_populates="users")


class ChargesFile(Base):
    __tablename__ = "invoices"
    id: Mapped[int] = mapped_column(primary_key=True)
    currency: Mapped[str] = mapped_column(String(3), nullable=False)
    amount: Mapped[Decimal | None] = mapped_column(sa.Numeric(18, 4), nullable=True)



from collections.abc import Callable
from typing import Any

from requela.builders.base import QueryBuilder


def get_builder_for_model(
    model: Any,
    resolve_alias_callback: Callable | None = None,
    validate_operator_and_field_callback: Callable | None = None,
    validate_ordering_callback: Callable | None = None,
):
    """Returns appropriate builder based on model type"""

    # SQLAlchemy model
    if hasattr(model, "__table__"):
        from requela.builders.sqlalchemy import SQLAlchemyQueryBuilder

        return SQLAlchemyQueryBuilder(
            model,
            resolve_alias_callback=resolve_alias_callback,
            validate_operator_and_field_callback=validate_operator_and_field_callback,
            validate_ordering_callback=validate_ordering_callback,
        )
    # Django model
    elif hasattr(model, "_meta"):  # pragma: no branch
        from requela.builders.django import DjangoQueryBuilder

        return DjangoQueryBuilder(
            model,
            resolve_alias_callback=resolve_alias_callback,
            validate_operator_and_field_callback=validate_operator_and_field_callback,
            validate_ordering_callback=validate_ordering_callback,
        )
    else:  # pragma: no cover
        raise ValueError(f"Unsupported model type: {type(model)}")


__all__ = ["get_builder_for_model", "QueryBuilder"]



from requela.builders import QueryBuilder, get_builder_for_model
from requela.dataclasses import Operator
from requela.rules import FieldRule, ModelRQLRules, RelationshipRule

__all__ = [
    "FieldRule",
    "ModelRQLRules",
    "Operator",
    "QueryBuilder",
    "get_builder_for_model",
    "RelationshipRule",
]



import enum

from django.db import models


class UserRole(str, enum.Enum):
    ADMIN = "admin"
    USER = "user"
    GUEST = "guest"


class AccountStatus(int, enum.Enum):
    ACTIVE = 1
    INACTIVE = 0
    PENDING = 2


class Account(models.Model):
    class Meta:
        db_table = "accounts"

    id = models.AutoField(primary_key=True)
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255, null=True)
    status = models.IntegerField(choices=[(status.value, status.name) for status in AccountStatus])
    balance = models.FloatField()
    created_at = models.DateTimeField()


class User(models.Model):
    class Meta:
        db_table = "users"

    id = models.AutoField(primary_key=True)
    name = models.CharField(max_length=255)
    age = models.IntegerField()
    email = models.CharField(max_length=255, null=True)
    role = models.CharField(max_length=255, choices=[(role.value, role.name) for role in UserRole])
    is_active = models.BooleanField()
    birth_date = models.DateField()
    account = models.ForeignKey(Account, on_delete=models.CASCADE, related_name="users")



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, IntEnum, StrEnum
from typing import Any, ClassVar

from requela.builders import QueryBuilder, get_builder_for_model
from requela.dataclasses import DEFAULT_OPERATORS, Operator

DOCS_HEADER = [
    "| Field | Operators | Order By |",
    "|-------|-----------|----------|",
]


@dataclass
class FieldRule:
    allowed_operators: set[Operator] | None = None
    alias: str | None = None
    allow_ordering: bool = True


@dataclass
class RelationshipRule:
    rules: ModelRQLRules
    alias: str | None = None


class ModelRQLRules:
    __model__: ClassVar[Any]
    _fields: ClassVar[dict[str, FieldRule]]
    _relations: ClassVar[dict[str, RelationshipRule]]

    def __init_subclass__(cls):
        super().__init_subclass__()

        if not hasattr(cls, "__model__"):
            raise TypeError(f"{cls.__name__} must define __model__")

        fields = {}
        relations = {}

        for attr_name in dir(cls):
            attr_value = getattr(cls, attr_name)

            if isinstance(attr_value, FieldRule):
                fields[attr_name] = attr_value
            elif isinstance(attr_value, RelationshipRule):
                relations[attr_name] = attr_value

        cls._fields = fields
        cls._relations = relations

    def __init__(self):
        self.builder = self._get_builder()
        self._validate()

    def build_query(self, rql_expression: str, initial_query: Any = None) -> Any:
        """Builds the query using the configured builder"""
        return self.builder.build_query(
            rql_expression,
            initial_query=initial_query,
        )

    def get_documentation(self) -> str:
        """Returns the documentation for the rules"""
        docs = DOCS_HEADER + self._get_fields_documentation()
        return "\n".join(docs)

    def _get_fields_documentation(self, alias_prefix: str = "") -> list[str]:
        """Returns the documentation for the rules"""

        docs = []
        for field_name, field in self._fields.items():
            operators = ", ".join(sorted([op.value for op in field.allowed_operators]))  # type: ignore
            prefix = ""
            if alias_prefix:
                prefix = f"{alias_prefix}."
            docs.append(
                f"|{prefix}{field.alias or field_name}"
                f"|{operators}|{'yes' if field.allow_ordering else 'no'}|"
            )
        for relation_name, relation in self._relations.items():
            prefix = ""
            if alias_prefix:
                prefix = f"{alias_prefix}."
            docs.append(f"|{prefix}{relation.alias or relation_name}|eq, ne|no|")
            docs.extend(
                relation.rules._get_fields_documentation(
                    f"{prefix}{relation.alias or relation_name}"
                )
            )
        return sorted(docs)

    @classmethod
    def _resolve_alias(cls, alias: str) -> str:
        field_name, _ = cls._get_field_by_alias(alias)
        return field_name

    @classmethod
    def _validate_operator_and_field(cls, field: str, operator: Operator):
        _, field_def = cls._get_field_by_alias(field)
        allowed_operators = (
            field_def.allowed_operators
            if isinstance(field_def, FieldRule)
            else {Operator.EQ, Operator.NE}
        )
        if operator not in allowed_operators:  # type: ignore
            raise ValueError(f"Operator '{operator.value}' is not allowed for field '{field}'.")

    @classmethod
    def _validate_ordering(cls, field: str):
        _, field_def = cls._get_field_by_alias(field)
        if not field_def.allow_ordering:  # type: ignore
            raise ValueError(f"Order by '{field}' is not allowed.")

    @classmethod
    def _get_builder(cls) -> QueryBuilder:
        return get_builder_for_model(
            cls.__model__,
            resolve_alias_callback=cls._resolve_alias,
            validate_operator_and_field_callback=cls._validate_operator_and_field,
            validate_ordering_callback=cls._validate_ordering,
        )

    def _validate(self) -> None:
        errors = []

        for field_name, field_def in self._fields.items():
            try:
                # Get field type from model (implementation depends on ORM)
                field_type = self._get_field_type(field_name)

                # If no operators specified, infer from field type
                if field_def.allowed_operators is None:
                    if issubclass(field_type, Enum | StrEnum | IntEnum):
                        default_operators = DEFAULT_OPERATORS.get(Enum)
                    else:
                        default_operators = DEFAULT_OPERATORS.get(field_type)

                    if default_operators is None:  # pragma: no cover
                        errors.append(
                            f"Cannot infer default operators for field {field_name} "
                            f"of type {field_type}"
                        )
                    else:
                        field_def.allowed_operators = default_operators

                # Validate that the operators are compatible with the field type
                if field_def.allowed_operators:  # pragma: no branch
                    invalid_ops = self._validate_operators(field_type, field_def.allowed_operators)
                    if invalid_ops:
                        invalid_ops_str = ", ".join(sorted([f"'{op.value}'" for op in invalid_ops]))
                        errors.append(
                            f"Invalid operators {invalid_ops_str} for field '{field_name}' "
                            f"of type '{field_type.__name__}'."
                        )

            except AttributeError:
                errors.append(
                    f"Field '{field_name}' not found in model '{self.__model__.__name__}'."
                )

        if errors:
            raise ExceptionGroup(
                f"Model validation failed for '{self.__model__.__name__}'",
                [ValueError(error_msg) for error_msg in errors],
            )

    def _get_field_type(self, field_name: str) -> type:
        """Gets the type of a field - to be implemented by specific ORM builders"""
        return self.builder.get_field_type(field_name)

    @classmethod
    def _validate_operators(cls, field_type: type, operators: set[Operator]) -> set[Operator]:
        """Validates operators against field type"""

        if issubclass(field_type, Enum | StrEnum | IntEnum):
            valid_operators = DEFAULT_OPERATORS.get(Enum)
        else:
            valid_operators = DEFAULT_OPERATORS.get(field_type)

        if not valid_operators:  # pragma: no cover
            return operators

        return {op for op in operators if op not in valid_operators}  # type: ignore

    @classmethod
    def _get_relation_by_alias(cls, alias: str) -> tuple[str, RelationshipRule]:
        relations = {}
        for relation_name, relation_def in cls._relations.items():
            relation_field_name = relation_def.alias or relation_name
            if alias.startswith(relation_field_name):
                relations[relation_name] = relation_def
        if not relations:
            raise ValueError(f"Relation with alias '{alias}' not found")
        if len(relations) > 1:
            raise ValueError(f"Multiple relations found for alias '{alias}'")
        return list(relations.items())[0]

    @classmethod
    def _get_field_by_alias(cls, alias: str) -> tuple[str, FieldRule | RelationshipRule]:
        """Gets a field definition by its alias or field name"""
        # First, check local fields
        for field_name, field_def in cls._fields.items():
            if field_def.alias == alias or field_name == alias:
                return field_name, field_def

        relation_name, relation_def = cls._get_relation_by_alias(alias)
        relation_field_name = relation_def.alias or relation_name
        if alias == relation_field_name:
            return relation_name, relation_def
        field_to_search = alias.removeprefix(relation_field_name)[1:] or relation_field_name
        f_name, f_def = relation_def.rules._get_field_by_alias(field_to_search)
        return f"{relation_name}.{f_name}", f_def



from requela.dataclasses import Operator
from requela.rules import FieldRule, ModelRQLRules, RelationshipRule
from tests.sqlalchemy.models import Account, Actor, Tenant, User


class NameMixin:
    name = FieldRule()


class ActorRules(NameMixin, ModelRQLRules):
    __model__ = Actor


class TimestampMixin:
    created_at = FieldRule(
        alias="events.created.at",
    )


class AuditableMixin(TimestampMixin):
    created_by = RelationshipRule(
        alias="events.created.by",
        rules=ActorRules(),
    )


class TenantRules(ModelRQLRules):
    __model__ = Tenant

    name = FieldRule()


class AccountRules(ModelRQLRules, AuditableMixin, NameMixin):
    __model__ = Account

    description = FieldRule()
    status = FieldRule()
    balance = FieldRule()
    datasource_id = FieldRule()
    tenant = RelationshipRule(rules=TenantRules())


class UserRules(ModelRQLRules):
    __model__ = User

    name = FieldRule()
    role = FieldRule(
        allowed_operators=[Operator.IN, Operator.OUT],
    )
    is_active = FieldRule()
    birth_date = FieldRule(
        alias="events.born.at",
    )
    account = RelationshipRule(
        rules=AccountRules(),
    )



def pytest_sessionstart(session):
    import django
    from django.conf import settings

    settings.configure(
        INSTALLED_APPS=[
            "tests.django",
        ],
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            }
        },
    )
    django.setup()



from sqlalchemy import select

from requela.builders.sqlalchemy import SQLAlchemyQueryBuilder
from tests.sqlalchemy.models import User
from tests.sqlalchemy.utils import assert_statements_equal


def test_simple_order_by():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "order_by(name)"
    stmt = builder.build_query(query_string)
    expected = select(User).order_by(User.name)
    assert_statements_equal(stmt, expected)


def test_simple_order_by_desc():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "order_by(-name)"
    stmt = builder.build_query(query_string)
    expected = select(User).order_by(User.name.desc())
    assert_statements_equal(stmt, expected)


def test_order_by_multiple_fields():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "order_by(name,age)"
    stmt = builder.build_query(query_string)
    expected = select(User).order_by(User.name, User.age)
    assert_statements_equal(stmt, expected)


def test_order_by_multiple_fields_desc():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "order_by(+name,-age)"
    stmt = builder.build_query(query_string)
    expected = select(User).order_by(User.name, User.age.desc())
    assert_statements_equal(stmt, expected)


def test_order_by_with_filter():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "order_by(name)&eq(age,30)"
    stmt = builder.build_query(query_string)
    expected = select(User).filter(User.age == 30).order_by(User.name)
    assert_statements_equal(stmt, expected)



from datetime import date, datetime

import pytest
from sqlalchemy import select
from sqlalchemy.orm import aliased

from requela.dataclasses import Operator
from requela.rules import FieldRule, ModelRQLRules
from tests.sqlalchemy.models import Account, Actor, User
from tests.sqlalchemy.rules import AccountRules, UserRules
from tests.sqlalchemy.utils import assert_statements_equal


def test_valid_field():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(name,Ratatouille 123)")
    assert_statements_equal(stmt, select(User).filter(User.name == "Ratatouille 123"))


def test_with_uuid():
    myfilter = AccountRules()
    stmt = myfilter.build_query("eq(datasource_id,90575008-bdde-4a40-ab07-82be547674e6)")
    assert_statements_equal(
        stmt,
        select(Account).filter(Account.datasource_id == "90575008-bdde-4a40-ab07-82be547674e6"),
    )


def test_valid_aliased_field():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(events.born.at,2024-01-01)")
    assert_statements_equal(stmt, select(User).filter(User.birth_date == date(2024, 1, 1)))


def test_invalid_field():
    user_filter = UserRules()
    with pytest.raises(ValueError):
        user_filter.build_query("eq(email,fail@example.com)")


def test_invalid_field_with_dot_notation():
    user_filter = UserRules()
    with pytest.raises(ValueError):
        user_filter.build_query("eq(norelationship.email,fail@example.com)")


def test_invalid_operator_for_field():
    user_filter = UserRules()
    with pytest.raises(ValueError, match="Operator 'eq' is not allowed for field 'role'."):
        user_filter.build_query("eq(role,admin)")


def test_valid_relationship():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(account.name,My Account)&order_by(name)")
    alias = aliased(Account)
    expected = select(User).join(alias).filter(alias.name == "My Account").order_by(User.name)
    assert_statements_equal(stmt, expected)


def test_valid_alias_on_relationship():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(account.events.created.at,2025-01-01T12:10:00+00:00)")
    alias = aliased(Account)
    expected = (
        select(User)
        .join(alias)
        .filter(alias.created_at == datetime.fromisoformat("2025-01-01T12:10:00+00:00"))
    )
    assert_statements_equal(stmt, expected)


def test_valid_aliased_relationship():
    account_filter = AccountRules()
    stmt = account_filter.build_query("eq(events.created.by.name,John Doe)")
    alias = aliased(Actor)
    expected = (
        select(Account)
        .join(alias, onclause=alias.id == Account.created_by_id)
        .filter(alias.name == "John Doe")
    )
    assert_statements_equal(stmt, expected)


def test_filter_class_validation_no_model():
    with pytest.raises(TypeError, match="UserRules must define __model__"):

        class UserRules(ModelRQLRules):
            pass


def test_filter_class_invalid_operator_for_field():
    class UserRules(ModelRQLRules):
        __model__ = User
        role = FieldRule(allowed_operators=[Operator.LTE, Operator.GTE])

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Invalid operators 'gte', 'lte' for field 'role' of type 'UserRole'."),
    ]


def test_filter_class_unexisting_field():
    class UserRules(ModelRQLRules):
        __model__ = User
        banana = FieldRule()

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Field 'banana' not found in model 'User'."),
    ]


def test_multiple_exceptions_raised():
    class UserRules(ModelRQLRules):
        __model__ = User
        role = FieldRule(allowed_operators=[Operator.LTE, Operator.GTE])
        banana = FieldRule()

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Field 'banana' not found in model 'User'."),
        (ValueError, "Invalid operators 'gte', 'lte' for field 'role' of type 'UserRole'."),
    ]


def test_filter_class_order_field_not_allowed():
    class UserRules(ModelRQLRules):
        __model__ = User
        name = FieldRule(allow_ordering=False)

    with pytest.raises(ValueError, match="Order by 'name' is not allowed."):
        UserRules().build_query("order_by(name)")


def test_get_documentation():
    docs = UserRules().get_documentation()
    assert docs.split("\n") == [
        "| Field | Operators | Order By |",
        "|-------|-----------|----------|",
        "|account.balance|eq, gt, gte, in, lt, lte, ne, out|yes|",
        "|account.datasource_id|eq, ilike, in, like, ne, out|yes|",
        "|account.description|eq, ilike, in, like, ne, out|yes|",
        "|account.events.created.at|eq, gt, gte, lt, lte, ne|yes|",
        "|account.events.created.by.name|eq, ilike, in, like, ne, out|yes|",
        "|account.events.created.by|eq, ne|no|",
        "|account.name|eq, ilike, in, like, ne, out|yes|",
        "|account.status|eq, in, ne, out|yes|",
        "|account.tenant.name|eq, ilike, in, like, ne, out|yes|",
        "|account.tenant|eq, ne|no|",
        "|account|eq, ne|no|",
        "|events.born.at|eq, gt, gte, lt, lte, ne|yes|",
        "|is_active|eq, ne|yes|",
        "|name|eq, ilike, in, like, ne, out|yes|",
        "|role|in, out|yes|",
    ]


def test_tenant_relationship_eq_null():
    account_rules = AccountRules()
    stmt = account_rules.build_query("eq(tenant,null())")
    assert_statements_equal(stmt, select(Account).filter(Account.tenant_id.is_(None)))


def test_tenant_relationship_eq_invalid_value():
    account_rules = AccountRules()
    with pytest.raises(
        ValueError, match="`eq` can be applied to relationship only to test for null."
    ):
        account_rules.build_query("eq(tenant,3)")


def test_tenant_relationship_ne_null():
    account_rules = AccountRules()
    stmt = account_rules.build_query("ne(tenant,null())")
    assert_statements_equal(stmt, select(Account).filter(Account.tenant_id.isnot(None)))


def test_tenant_relationship_ne_invalid_value():
    account_rules = AccountRules()
    with pytest.raises(
        ValueError, match="`ne` can be applied to relationship only to test for null."
    ):
        account_rules.build_query("ne(tenant,pip)")



from datetime import date, datetime

from lark import Transformer

from requela.dataclasses import (
    FilterExpression,
    OperatorFunctions,
    OrderByExpression,
    OrderField,
)


class RQLTransformer(Transformer):
    def __init__(self, operators: OperatorFunctions):
        super().__init__()
        self.operators = operators

    def start(self, args):
        return args[0]

    def expression(self, args):
        # Extract the actual operation from the expression node
        return args[0]

    def logical_expression(self, args):
        # apply the logical operator to the list of conditions
        operator_func = args[0]
        argument_list = args[1]
        conditions = [argument.condition for argument in argument_list]
        return FilterExpression(condition=operator_func(*conditions))

    def and_op(self, _):
        return self.operators.and_op

    def or_op(self, _):
        return self.operators.or_op

    def not_op(self, _):
        return self.operators.not_op

    def eq_op(self, _):
        return self.operators.eq_op

    def ne_op(self, _):
        return self.operators.ne_op

    def gt_op(self, _):
        return self.operators.gt_op

    def lt_op(self, _):
        return self.operators.lt_op

    def gte_op(self, _):
        return self.operators.gte_op

    def lte_op(self, _):
        return self.operators.lte_op

    def in_op(self, _):
        return self.operators.in_op

    def out_op(self, _):
        return self.operators.out_op

    def like_op(self, _):
        return self.operators.like_op

    def ilike_op(self, _):
        return self.operators.ilike_op

    def argument_list(self, args):
        return list(args)

    def argument(self, args):
        return args[0]

    def comparison(self, args):
        operator, prop, value = args
        return FilterExpression(
            condition=operator(prop, value),
        )

    def property(self, args):
        return str(args[0])

    def value(self, args):
        # Directly return the processed literal
        return args[0]

    def literal(self, args):
        value = args[0]
        if isinstance(value, str):
            if value == "true":
                return True
            if value == "false":
                return False
            if value == "null()":
                return None
            if value == "empty()":
                return ""
        return value

    def FLOAT(self, token):
        return float(token)

    def INT(self, token):
        return int(token)

    def DATETIME(self, token):
        return datetime.fromisoformat(token)

    def DATE(self, token):
        return date.fromisoformat(token)

    def UNQUOTED_VAL(self, token):
        return str(token)

    def UUID(self, token):
        return str(token)

    def tuple(self, args):
        return tuple(args)

    def any_expression(self, args):
        relationship_name, filter_expression = args
        return FilterExpression(
            condition=self.operators.any_op(relationship_name, filter_expression.condition)
        )

    def order_expression(self, args):
        order_list = args[0]
        return OrderByExpression(fields=order_list)

    def order_list(self, args):
        return list(args)

    def order_item(self, args):
        return OrderField(direction=args[0], field_path=args[1])

    def SIGN(self, token):
        return str(token)

    def combined_expression(self, args):
        # Handle multiple operations
        operations = []
        for arg in args:
            operations.append(arg)
        return operations

    def single_expression(self, args):
        # Extract the actual operation from the single_expression node
        return args[0]



from datetime import date, datetime

import pytest
from sqlalchemy import select

from requela.builders.sqlalchemy import SQLAlchemyQueryBuilder
from tests.sqlalchemy.models import Account, ChargesFile, User
from tests.sqlalchemy.utils import assert_statements_equal


# Equality tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "name", "Ratatouille 123", select(User).filter(User.name == "Ratatouille 123")),
        (User, "age", 25, select(User).filter(User.age == 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance == 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance == -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance == 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date == date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at == datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
        (User, "is_active", "true", select(User).filter(User.is_active.is_(True))),
        (User, "is_active", "false", select(User).filter(User.is_active.is_(False))),
        (User, "name", "empty()", select(User).filter(User.name == "")),
        (User, "name", "null()", select(User).filter(User.name.is_(None))),
    ],
)
def test_comparison_eq(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"eq({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Not equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "name", "Ratatouille", select(User).filter(User.name != "Ratatouille")),
        (User, "age", 25, select(User).filter(User.age != 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance != 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance != -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance != 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date != date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at != datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
        (User, "is_active", "true", select(User).filter(User.is_active.isnot(True))),
        (User, "is_active", "false", select(User).filter(User.is_active.isnot(False))),
        (User, "name", "empty()", select(User).filter(User.name != "")),
        (User, "name", "null()", select(User).filter(User.name.isnot(None))),
    ],
)
def test_comparison_ne(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"ne({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Less than tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, select(User).filter(User.age < 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance < 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance < -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance < 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date < date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at < datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_lt(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"lt({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Greater than tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, select(User).filter(User.age > 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance > 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance > -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance > 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date > date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at > datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_gt(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"gt({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Less than or equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, select(User).filter(User.age <= 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance <= 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance <= -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance <= 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date <= date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at <= datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_lte(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"lte({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Greater than or equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, select(User).filter(User.age >= 25)),
        (Account, "balance", 25.13, select(Account).filter(Account.balance >= 25.13)),
        (Account, "balance", -71.14, select(Account).filter(Account.balance >= -71.14)),
        (Account, "balance", 33, select(Account).filter(Account.balance >= 33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            select(User).filter(User.birth_date >= date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            select(Account).filter(
                Account.created_at >= datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_gte(model, field, value, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"gte({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# In tests
@pytest.mark.parametrize(
    ("model", "field", "values", "expected"),
    [
        (User, "age", "(25,30,35)", select(User).filter(User.age.in_([25, 30, 35]))),
        (
            Account,
            "balance",
            "(25.13,50.00,75.25)",
            select(Account).filter(Account.balance.in_([25.13, 50.00, 75.25])),
        ),
        (
            User,
            "name",
            "(Alice,Bob,Charlie)",
            select(User).filter(User.name.in_(["Alice", "Bob", "Charlie"])),
        ),
    ],
)
def test_comparison_in(model, field, values, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"in({field},{values})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Out tests
@pytest.mark.parametrize(
    ("model", "field", "values", "expected"),
    [
        (User, "age", "(25,30,35)", select(User).filter(User.age.not_in([25, 30, 35]))),
        (
            Account,
            "balance",
            "(25.13,50.00,75.25)",
            select(Account).filter(Account.balance.not_in([25.13, 50.00, 75.25])),
        ),
        (
            User,
            "name",
            "(Alice,Bob,Charlie)",
            select(User).filter(User.name.not_in(["Alice", "Bob", "Charlie"])),
        ),
    ],
)
def test_comparison_out(model, field, values, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"out({field},{values})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Like tests
@pytest.mark.parametrize(
    ("model", "field", "pattern", "expected"),
    [
        (User, "name", "Rat*", select(User).filter(User.name.like("Rat%"))),
        (User, "name", "*ouille", select(User).filter(User.name.like("%ouille"))),
        (User, "name", "*atato*", select(User).filter(User.name.like("%atato%"))),
    ],
)
def test_comparison_like(model, field, pattern, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"like({field},{pattern})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# ILike tests
@pytest.mark.parametrize(
    ("model", "field", "pattern", "expected"),
    [
        (User, "name", "rat*", select(User).filter(User.name.ilike("rat%"))),
        (User, "name", "*OUILLE", select(User).filter(User.name.ilike("%OUILLE"))),
        (User, "name", "*AtAtO*", select(User).filter(User.name.ilike("%AtAtO%"))),
    ],
)
def test_comparison_ilike(model, field, pattern, expected):
    builder = SQLAlchemyQueryBuilder(model)
    query_string = f"ilike({field},{pattern})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


def test_with_initial_query():
    builder = SQLAlchemyQueryBuilder(User)
    initial = select(User).filter(User.age > 25)
    query_string = "eq(name,Ratatouille)"
    stmt = builder.build_query(query_string, initial)
    assert_statements_equal(stmt, select(User).filter(User.age > 25, User.name == "Ratatouille"))


def test_decimal_field():
    builder = SQLAlchemyQueryBuilder(ChargesFile)
    query_string = "eq(amount,100.40)"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, select(ChargesFile).filter(ChargesFile.amount == 100.40))


def test_decimal_gt_field():
    builder = SQLAlchemyQueryBuilder(ChargesFile)
    query_string = "gt(amount,100.40)"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, select(ChargesFile).filter(ChargesFile.amount > 100.40))


def test_decimal_gte_field():
    builder = SQLAlchemyQueryBuilder(ChargesFile)
    query_string = "gte(amount,100.40)"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, select(ChargesFile).filter(ChargesFile.amount >= 100.40))


def test_decimal_lte_field():
    builder = SQLAlchemyQueryBuilder(ChargesFile)
    query_string = "lte(amount,100.40)"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, select(ChargesFile).filter(ChargesFile.amount <= 100.40))


def test_decimal_lt_field():
    builder = SQLAlchemyQueryBuilder(ChargesFile)
    query_string = "lt(amount,100.40)"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, select(ChargesFile).filter(ChargesFile.amount < 100.40))



from datetime import date, datetime

import pytest

from requela.dataclasses import Operator
from requela.rules import FieldRule, ModelRQLRules
from tests.django.models import User
from tests.django.rules import UserRules
from tests.django.utils import assert_statements_equal


def test_valid_field():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(name,Ratatouille 123)")
    assert_statements_equal(stmt, User.objects.filter(name="Ratatouille 123"))


def test_valid_aliased_field():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(events.born.at,2024-01-01)")
    assert_statements_equal(stmt, User.objects.filter(birth_date=date(2024, 1, 1)))


def test_invalid_field():
    user_filter = UserRules()
    with pytest.raises(ValueError):
        user_filter.build_query("eq(email,fail@example.com)")


def test_invalid_field_with_dot_notation():
    user_filter = UserRules()
    with pytest.raises(ValueError):
        user_filter.build_query("eq(norelationship.email,fail@example.com)")


def test_invalid_operator_for_field():
    user_filter = UserRules()
    with pytest.raises(ValueError, match="Operator 'eq' is not allowed for field 'role'."):
        user_filter.build_query("eq(role,admin)")


def test_valid_relationship():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(account.name,My Account)&order_by(name)")
    expected = User.objects.filter(account__name="My Account").order_by("name")
    assert_statements_equal(stmt, expected)


def test_valid_aliased_relationship():
    user_filter = UserRules()
    stmt = user_filter.build_query("eq(account.events.created.at,2025-01-01T12:10:00+00:00)")
    expected = User.objects.filter(
        account__created_at=datetime.fromisoformat("2025-01-01T12:10:00+00:00")
    )
    assert_statements_equal(stmt, expected)


def test_filter_class_validation_no_model():
    with pytest.raises(TypeError, match="UserRules must define __model__"):

        class UserRules(ModelRQLRules):
            pass


def test_filter_class_invalid_operator_for_field():
    class UserRules(ModelRQLRules):
        __model__ = User
        role = FieldRule(allowed_operators=[Operator.LTE, Operator.GTE])

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Invalid operators 'gte', 'lte' for field 'role' of type 'str'."),
    ]


def test_filter_class_unexisting_field():
    class UserRules(ModelRQLRules):
        __model__ = User
        banana = FieldRule()

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Field 'banana' not found in model 'User'."),
    ]


def test_multiple_exceptions_raised():
    class UserRules(ModelRQLRules):
        __model__ = User
        role = FieldRule(allowed_operators=[Operator.LTE, Operator.GTE])
        banana = FieldRule()

    with pytest.raises(ExceptionGroup, match="Model validation failed for 'User'") as exc:
        UserRules()

    assert [(e.__class__, str(e)) for e in exc.value.exceptions] == [
        (ValueError, "Field 'banana' not found in model 'User'."),
        (ValueError, "Invalid operators 'gte', 'lte' for field 'role' of type 'str'."),
    ]


def test_filter_class_order_field_not_allowed():
    class UserRules(ModelRQLRules):
        __model__ = User
        name = FieldRule(allow_ordering=False)

    with pytest.raises(ValueError, match="Order by 'name' is not allowed."):
        UserRules().build_query("order_by(name)")



from functools import lru_cache
from pathlib import Path

from lark import Lark, LarkError
from lark.tree import ParseTree

with open(Path(__file__).parent / "grammar.lark") as f:
    GRAMMAR = f.read()


PARSER = Lark(GRAMMAR, parser="lalr")


@lru_cache(maxsize=1000)
def parse(rql_query: str) -> ParseTree:
    try:
        return PARSER.parse(rql_query)
    except LarkError as e:
        raise ValueError(f"Invalid RQL query: {e}") from e



from requela.rules import FieldRule, ModelRQLRules, Operator, RelationshipRule
from tests.django.models import Account, User


class AccountRules(ModelRQLRules):
    __model__ = Account

    name = FieldRule()
    description = FieldRule()
    status = FieldRule()
    balance = FieldRule()
    created_at = FieldRule(
        alias="events.created.at",
    )


class UserRules(ModelRQLRules):
    __model__ = User

    name = FieldRule()
    role = FieldRule(
        allowed_operators=[Operator.IN, Operator.OUT],
    )
    is_active = FieldRule()
    birth_date = FieldRule(
        alias="events.born.at",
    )
    account = RelationshipRule(
        alias="account",
        rules=AccountRules(),
    )



from django.db.models import QuerySet


def assert_statements_equal(stmt: QuerySet, expected: QuerySet):
    assert stmt.query.sql_with_params() == expected.query.sql_with_params(), (
        f"\nSQL Parameters differ:\nActual:\n{stmt.query.sql_with_params()}\n"
        f"Expected:\n{expected.query.sql_with_params()}"
    )



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import date, datetime

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from django.db.models.query import QuerySet

from requela.builders.base import QueryBuilder
from requela.dataclasses import FilterExpression, OrderByExpression


@dataclass
class PrefetchExpression:
    relationship_name: str
    condition: Q


class DjangoQueryBuilder(QueryBuilder):
    def get_initial_query(self):
        return self.model_class.objects.all()

    def get_field_type(self, field: str) -> type:
        try:
            django_field = self.model_class._meta.get_field(field)
            if django_field.get_internal_type() == "CharField":
                return str
            elif django_field.get_internal_type() == "DateTimeField":
                return datetime
            elif django_field.get_internal_type() == "DateField":
                return date
            elif django_field.get_internal_type() == "IntegerField":
                return int
            elif django_field.get_internal_type() == "FloatField":
                return float
            elif django_field.get_internal_type() == "BooleanField":
                return bool
            else:
                raise ValueError(f"Unsupported field type: {django_field.get_internal_type()}")
        except FieldDoesNotExist:
            raise AttributeError(
                f"Field '{field}' not found in model '{self.model_class.__name__}'."
            )

    def apply_and(self, *conditions: Q) -> Q:
        query = Q()
        for condition in conditions:
            query &= condition
        return query

    def apply_or(self, *conditions: Q) -> Q:
        query = Q()
        for condition in conditions:
            query |= condition
        return query

    def apply_not(self, condition: Q) -> Q:
        return ~condition

    def apply_eq(self, prop: str, value: str | bool | date | datetime | int | float | None) -> Q:
        if value is None:
            return Q(**{f"{self.resolve_property(prop)}__isnull": True})
        return Q(**{self.resolve_property(prop): value})

    def apply_ne(self, prop: str, value: str | bool | date | datetime | int | float | None) -> Q:
        if value is None:
            return Q(**{f"{self.resolve_property(prop)}__isnull": False})
        return ~Q(**{self.resolve_property(prop): value})

    def apply_gt(self, prop: str, value: date | datetime | int | float) -> Q:
        return Q(**{f"{self.resolve_property(prop)}__gt": value})

    def apply_lt(self, prop: str, value: date | datetime | int | float) -> Q:
        return Q(**{f"{self.resolve_property(prop)}__lt": value})

    def apply_gte(self, prop: str, value: date | datetime | int | float) -> Q:
        return Q(**{f"{self.resolve_property(prop)}__gte": value})

    def apply_lte(self, prop: str, value: date | datetime | int | float) -> Q:
        return Q(**{f"{self.resolve_property(prop)}__lte": value})

    def apply_in(self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]) -> Q:
        return Q(**{f"{self.resolve_property(prop)}__in": value})

    def apply_out(self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]) -> Q:
        return ~Q(**{f"{self.resolve_property(prop)}__in": value})

    def apply_like(self, prop: str, value: str) -> Q:
        if value.startswith("*") and value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__contains": value[1:-1]})
        elif value.startswith("*") and not value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__endswith": value[1:]})
        elif not value.startswith("*") and value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__startswith": value[:-1]})
        else:
            return Q(**{f"{self.resolve_property(prop)}__contains": value})

    def apply_ilike(self, prop: str, value: str) -> Q:
        if value.startswith("*") and value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__icontains": value[1:-1]})
        elif value.startswith("*") and not value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__iendswith": value[1:]})
        elif not value.startswith("*") and value.endswith("*"):
            return Q(**{f"{self.resolve_property(prop)}__istartswith": value[:-1]})
        else:
            return Q(**{f"{self.resolve_property(prop)}__icontains": value})

    def resolve_property(self, prop_path: str) -> str:
        prop = self.resolve_alias(prop_path)
        return prop.replace(".", "__")

    def apply_any(self, relationship_name, condition: Q) -> PrefetchExpression:
        return PrefetchExpression(relationship_name, condition)

    def apply_filter(self, query: QuerySet, filter_expression: FilterExpression) -> QuerySet:
        if isinstance(filter_expression.condition, PrefetchExpression):
            query = query.prefetch_related(filter_expression.condition.relationship_name)
            return query.filter(filter_expression.condition.condition)
        return query.filter(filter_expression.condition)

    def apply_joins(self, query: QuerySet):
        return query

    def apply_order_by(self, query: QuerySet, order_by_expression: OrderByExpression) -> QuerySet:
        fields = []
        for order_field in order_by_expression.fields:
            prop = self.resolve_property(order_field.field_path)
            if order_field.direction == "-":
                prop = f"-{prop}"
            fields.append(prop)
        return query.order_by(*fields)



from sqlalchemy import exists, select
from sqlalchemy.orm import aliased

from requela.builders.sqlalchemy import SQLAlchemyQueryBuilder
from tests.sqlalchemy.models import Account, User
from tests.sqlalchemy.utils import assert_statements_equal


def test_comparison_eq_many_to_one():
    builder = SQLAlchemyQueryBuilder(User)
    query_string = "eq(account.name,My Account)"
    stmt = builder.build_query(query_string)
    alias = aliased(Account)
    expected = select(User).join(alias).filter(alias.name == "My Account")
    assert_statements_equal(stmt, expected)


def test_comparison_any_one_to_many():
    builder = SQLAlchemyQueryBuilder(Account)
    query_string = "any(users,eq(users.name,John))"
    stmt = builder.build_query(query_string)
    alias = aliased(User)
    expected = select(Account).filter(
        exists(alias).where(alias.account_id == Account.id, alias.name == "John")
    )
    assert_statements_equal(stmt, expected)


def test_comparison_any_one_to_many_complex_filter():
    builder = SQLAlchemyQueryBuilder(Account)
    query_string = "any(users,and(eq(users.name,John),gt(users.age,30)))"
    stmt = builder.build_query(query_string)
    alias = aliased(User)
    expected = select(Account).filter(
        exists(alias).where(
            alias.account_id == Account.id,
            alias.name == "John",
            alias.age > 30,
        )
    )
    assert_statements_equal(stmt, expected)



import pytest

from requela.parser import parse


def test_parser():
    ast = parse("eq(name,John)")
    assert ast is not None


def test_parser_invalid():
    with pytest.raises(ValueError, match="Invalid RQL query: Unexpected token"):
        parse("eq(name,John")



from django.db.models import Q

from requela.builders.django import DjangoQueryBuilder
from tests.django.models import Account, User
from tests.django.utils import assert_statements_equal


def test_comparison_eq_many_to_one():
    builder = DjangoQueryBuilder(User)
    query_string = "eq(account__name,My Account)"
    stmt = builder.build_query(query_string)
    expected = User.objects.filter(account__name="My Account")
    assert_statements_equal(stmt, expected)


def test_comparison_any_one_to_many():
    builder = DjangoQueryBuilder(Account)
    query_string = "any(users,eq(users.name,John))"
    stmt = builder.build_query(query_string)
    expected = Account.objects.prefetch_related("users").filter(users__name="John")
    assert_statements_equal(stmt, expected)


def test_comparison_any_one_to_many_complex_filter():
    builder = DjangoQueryBuilder(Account)
    query_string = "any(users,and(eq(users.name,John),gt(users.age,30)))"
    stmt = builder.build_query(query_string)
    expected = Account.objects.prefetch_related("users").filter(
        Q(users__name="John") & Q(users__age__gt=30)
    )
    assert_statements_equal(stmt, expected)



import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from datetime import date, datetime
from functools import partial
from typing import Any

from lark.exceptions import VisitError

from requela.dataclasses import FilterExpression, Operator, OperatorFunctions, OrderByExpression
from requela.parser import parse
from requela.transformer import RQLTransformer

logger = logging.getLogger(__name__)


class QueryBuilder(ABC):
    def __init__(
        self,
        model_class: Any,
        resolve_alias_callback: Callable | None = None,
        validate_operator_and_field_callback: Callable | None = None,
        validate_ordering_callback: Callable | None = None,
    ):
        self.model_class = model_class
        self.resolve_alias_callback = resolve_alias_callback
        self.validate_operator_and_field_callback = validate_operator_and_field_callback
        self.validate_ordering_callback = validate_ordering_callback
        self.transformer = RQLTransformer(
            OperatorFunctions(
                and_op=self.apply_and,
                or_op=self.apply_or,
                not_op=self.apply_not,
                eq_op=partial(self.apply_operator, Operator.EQ),
                ne_op=partial(self.apply_operator, Operator.NE),
                gt_op=partial(self.apply_operator, Operator.GT),
                lt_op=partial(self.apply_operator, Operator.LT),
                gte_op=partial(self.apply_operator, Operator.GTE),
                lte_op=partial(self.apply_operator, Operator.LTE),
                in_op=partial(self.apply_operator, Operator.IN),
                out_op=partial(self.apply_operator, Operator.OUT),
                like_op=partial(self.apply_operator, Operator.LIKE),
                ilike_op=partial(self.apply_operator, Operator.ILIKE),
                any_op=self.apply_any,
            )
        )

    @abstractmethod
    def get_initial_query(self):
        pass

    @abstractmethod
    def get_field_type(self, field: str) -> type:
        pass

    @abstractmethod
    def apply_and(self, *conditions):
        pass

    @abstractmethod
    def apply_or(self, *conditions):
        pass

    @abstractmethod
    def apply_not(self, condition):
        pass

    @abstractmethod
    def apply_eq(self, prop: str, value: str | date | datetime | int | float | None):
        pass

    @abstractmethod
    def apply_ne(self, prop: str, value: str | date | datetime | int | float | None):
        pass

    @abstractmethod
    def apply_gt(self, prop: str, value: date | datetime | int | float):
        pass

    @abstractmethod
    def apply_lt(self, prop: str, value: date | datetime | int | float):
        pass

    @abstractmethod
    def apply_gte(self, prop: str, value: date | datetime | int | float):
        pass

    @abstractmethod
    def apply_lte(self, prop: str, value: date | datetime | int | float):
        pass

    @abstractmethod
    def apply_in(self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]):
        pass

    @abstractmethod
    def apply_out(self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]):
        pass

    @abstractmethod
    def apply_like(self, prop: str, value: str):
        pass

    @abstractmethod
    def apply_ilike(self, prop: str, value: str):
        pass

    @abstractmethod
    def apply_any(self, prop: str, condition: Any) -> Any:
        pass

    @abstractmethod
    def apply_joins(self, query):
        pass

    @abstractmethod
    def apply_order_by(self, query: Any, order_by_expression: OrderByExpression) -> Any:
        pass

    @abstractmethod
    def apply_filter(self, query, filter_expression: FilterExpression) -> Any:
        pass

    def apply_operator(self, operator: Operator, prop: str, value: Any):
        self.validate_operator_and_field(prop, operator)
        return getattr(self, f"apply_{operator.value}")(prop, value)

    def validate_operator_and_field(self, field: str, operator: Operator) -> None:
        if self.validate_operator_and_field_callback:
            return self.validate_operator_and_field_callback(field, operator)

    def resolve_alias(self, alias: str) -> str:
        if self.resolve_alias_callback:
            return self.resolve_alias_callback(alias)
        return alias

    def build_query(self, rql_query: str, initial_query: Any = None) -> Any:
        query = initial_query if initial_query is not None else self.get_initial_query()
        ast = parse(rql_query)
        try:
            expressions = self.transformer.transform(ast)
        except VisitError as e:
            raise e.orig_exc
        for expression in expressions:
            if isinstance(expression, FilterExpression):
                query = self.apply_filter(query, expression)
            if isinstance(expression, OrderByExpression):
                if self.validate_ordering_callback:
                    for field in expression.fields:
                        self.validate_ordering_callback(field.field_path)
                query = self.apply_order_by(query, expression)
        return self.apply_joins(query)



from collections.abc import Callable
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Any


@dataclass
class FilterExpression:
    condition: Any


@dataclass
class OrderField:
    direction: str
    field_path: str


@dataclass
class JoinExpression:
    target: Any
    on: Any
    is_outer: bool = False


@dataclass
class OrderByExpression:
    fields: list[OrderField]


class Operator(Enum):
    EQ = "eq"
    NE = "ne"
    GT = "gt"
    LT = "lt"
    GTE = "gte"
    LTE = "lte"
    IN = "in"
    OUT = "out"
    LIKE = "like"
    ILIKE = "ilike"
    ANY = "any"


# Default operator sets based on field types
DEFAULT_OPERATORS = {
    str: {
        Operator.EQ,
        Operator.NE,
        Operator.IN,
        Operator.OUT,
        Operator.LIKE,
        Operator.ILIKE,
    },
    int: {
        Operator.EQ,
        Operator.NE,
        Operator.GT,
        Operator.LT,
        Operator.GTE,
        Operator.LTE,
        Operator.IN,
        Operator.OUT,
    },
    float: {
        Operator.EQ,
        Operator.NE,
        Operator.GT,
        Operator.LT,
        Operator.GTE,
        Operator.LTE,
        Operator.IN,
        Operator.OUT,
    },
    bool: {Operator.EQ, Operator.NE},
    datetime: {
        Operator.EQ,
        Operator.NE,
        Operator.GT,
        Operator.LT,
        Operator.GTE,
        Operator.LTE,
    },
    date: {
        Operator.EQ,
        Operator.NE,
        Operator.GT,
        Operator.LT,
        Operator.GTE,
        Operator.LTE,
    },
    Enum: {Operator.EQ, Operator.NE, Operator.IN, Operator.OUT},
    Decimal: {
        Operator.EQ,
        Operator.NE,
        Operator.GT,
        Operator.LT,
        Operator.GTE,
        Operator.LTE,
        Operator.IN,
        Operator.OUT,
    },
}


@dataclass
class OperatorFunctions:
    # logical
    and_op: Callable
    or_op: Callable
    not_op: Callable

    # comparison
    eq_op: Callable
    ne_op: Callable
    gt_op: Callable
    lt_op: Callable
    gte_op: Callable
    lte_op: Callable
    in_op: Callable
    out_op: Callable
    like_op: Callable
    ilike_op: Callable

    # any operator
    any_op: Callable



import pytest
from sqlalchemy import and_, or_, select

from requela.builders.sqlalchemy import SQLAlchemyQueryBuilder
from tests.sqlalchemy.models import User
from tests.sqlalchemy.utils import assert_statements_equal


@pytest.mark.parametrize(
    ("query_string", "expected"),
    [
        ("and(eq(name,John),eq(age,30))", select(User).filter(User.name == "John", User.age == 30)),
        (
            "and(eq(name,John),eq(age,30),eq(email,test@example.com))",
            select(User).filter(
                User.name == "John", User.age == 30, User.email == "test@example.com"
            ),
        ),
    ],
)
def test_logical_and(query_string, expected):
    builder = SQLAlchemyQueryBuilder(User)
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


@pytest.mark.parametrize(
    ("query_string", "expected"),
    [
        (
            "or(eq(name,John),eq(age,30))",
            select(User).filter(or_(User.name == "John", User.age == 30)),
        ),
        (
            "or(eq(name,John),eq(age,30),eq(email,test@example.com))",
            select(User).filter(
                or_(User.name == "John", User.age == 30, User.email == "test@example.com")
            ),
        ),
    ],
)
def test_logical_or(query_string, expected):
    builder = SQLAlchemyQueryBuilder(User)
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


def test_logical_not():
    builder = SQLAlchemyQueryBuilder(User)
    stmt = builder.build_query("not(eq(name,John))")
    assert_statements_equal(stmt, select(User).filter(User.name != "John"))


def test_nested_logical():
    builder = SQLAlchemyQueryBuilder(User)
    stmt = builder.build_query("and(or(eq(name,John),eq(age,30)),eq(email,test@example.com))")
    assert_statements_equal(
        stmt,
        select(User).filter(
            and_(or_(User.name == "John", User.age == 30), User.email == "test@example.com")
        ),
    )


def test_nested_logical_with_not():
    builder = SQLAlchemyQueryBuilder(User)
    stmt = builder.build_query(
        "and(or(eq(name,John),eq(age,30)),not(in(email,(test@example.com,test2@example.com))))"
    )
    assert_statements_equal(
        stmt,
        select(User).filter(
            and_(
                or_(User.name == "John", User.age == 30),
                User.email.notin_(["test@example.com", "test2@example.com"]),
            )
        ),
    )



def assert_statements_equal(stmt, expected):
    compiled = stmt.compile()
    expected_compiled = expected.compile()
    assert str(compiled.string) == str(expected_compiled.string), (
        f"\nSQL Statement differs:\nActual:\n{compiled.string}\n"
        f"Expected:\n{expected_compiled.string}"
    )

    # Compare parameters with better formatting for pytest output
    assert compiled.params == expected_compiled.params, (
        f"\nSQL Parameters differ:\nActual:\n{compiled.params}\n"
        f"Expected:\n{expected_compiled.params}"
    )



from requela.builders.django import DjangoQueryBuilder
from tests.django.models import User
from tests.django.utils import assert_statements_equal


def test_simple_order_by():
    builder = DjangoQueryBuilder(User)
    query_string = "order_by(name)"
    stmt = builder.build_query(query_string)
    expected = User.objects.order_by("name")
    assert_statements_equal(stmt, expected)


def test_simple_order_by_desc():
    builder = DjangoQueryBuilder(User)
    query_string = "order_by(-name)"
    stmt = builder.build_query(query_string)
    expected = User.objects.order_by("-name")
    assert_statements_equal(stmt, expected)


def test_order_by_multiple_fields():
    builder = DjangoQueryBuilder(User)
    query_string = "order_by(name,age)"
    stmt = builder.build_query(query_string)
    expected = User.objects.order_by("name", "age")
    assert_statements_equal(stmt, expected)


def test_order_by_multiple_fields_desc():
    builder = DjangoQueryBuilder(User)
    query_string = "order_by(+name,-age)"
    stmt = builder.build_query(query_string)
    expected = User.objects.order_by("name", "-age")
    assert_statements_equal(stmt, expected)


def test_order_by_with_filter():
    builder = DjangoQueryBuilder(User)
    query_string = "order_by(name)&eq(age,30)"
    stmt = builder.build_query(query_string)
    expected = User.objects.filter(age=30).order_by("name")
    assert_statements_equal(stmt, expected)



from collections.abc import Callable, Sequence
from datetime import date, datetime

from sqlalchemy import (
    BooleanClauseList,
    ColumnElement,
    ColumnExpressionArgument,
    Exists,
    UnaryExpression,
    and_,
    exists,
    or_,
    select,
)
from sqlalchemy.orm import DeclarativeBase, Query, RelationshipProperty, aliased

from requela.builders.base import QueryBuilder
from requela.dataclasses import FilterExpression, JoinExpression, OrderByExpression


class SQLAlchemyQueryBuilder(QueryBuilder):
    def __init__(
        self,
        model_class: DeclarativeBase,
        resolve_alias_callback: Callable | None = None,
        validate_operator_and_field_callback: Callable | None = None,
        validate_ordering_callback: Callable | None = None,
    ):
        super().__init__(
            model_class,
            resolve_alias_callback=resolve_alias_callback,
            validate_operator_and_field_callback=validate_operator_and_field_callback,
            validate_ordering_callback=validate_ordering_callback,
        )
        self.joins: list[JoinExpression] = []

    def get_initial_query(self):
        return select(self.model_class)

    def get_field_type(self, field: str) -> type:
        field_type = getattr(self.model_class, field).property.columns[0].type.python_type
        return field_type

    def apply_and(self, *conditions: ColumnExpressionArgument) -> ColumnElement:
        return and_(*conditions)

    def apply_or(self, *conditions: ColumnExpressionArgument) -> ColumnElement:
        return or_(*conditions)

    def apply_not(self, condition: ColumnElement) -> ColumnElement:
        return ~condition

    def apply_eq(
        self, prop: str, value: str | bool | date | datetime | int | float | None
    ) -> ColumnExpressionArgument:
        model_field = self.resolve_property(prop)
        if isinstance(model_field.property, RelationshipProperty):
            if value is not None:
                raise ValueError("`eq` can be applied to relationship only to test for null.")
            return self.apply_eq_to_relationship(model_field.property)

        if value is True or value is False or value is None:
            return model_field.is_(value)
        return model_field == value

    def apply_eq_to_relationship(
        self, relationship_property: RelationshipProperty
    ) -> ColumnExpressionArgument:
        conditions = []
        for local_col in relationship_property.local_columns:
            conditions.append(local_col.is_(None))
        return and_(*conditions)

    def apply_ne(
        self, prop: str, value: str | bool | date | datetime | int | float | None
    ) -> ColumnExpressionArgument:
        model_field = self.resolve_property(prop)
        if isinstance(model_field.property, RelationshipProperty):
            if value is not None:
                raise ValueError("`ne` can be applied to relationship only to test for null.")
            return self.apply_ne_to_relationship(model_field.property)
        if value is True or value is False or value is None:
            return model_field.isnot(value)
        return model_field != value

    def apply_ne_to_relationship(
        self, relationship_property: RelationshipProperty
    ) -> ColumnExpressionArgument:
        conditions = []
        for local_col in relationship_property.local_columns:
            conditions.append(local_col.isnot(None))
        return or_(*conditions)

    def apply_gt(self, prop: str, value: date | datetime | int | float) -> ColumnExpressionArgument:
        return self.resolve_property(prop) > value

    def apply_lt(self, prop: str, value: date | datetime | int | float) -> ColumnExpressionArgument:
        return self.resolve_property(prop) < value

    def apply_gte(
        self, prop: str, value: date | datetime | int | float
    ) -> ColumnExpressionArgument:
        return self.resolve_property(prop) >= value

    def apply_lte(
        self, prop: str, value: date | datetime | int | float
    ) -> ColumnExpressionArgument:
        return self.resolve_property(prop) <= value

    def apply_in(
        self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]
    ) -> ColumnExpressionArgument:
        return self.resolve_property(prop).in_(value)

    def apply_out(
        self, prop: str, value: Sequence[str] | Sequence[float] | Sequence[int]
    ) -> ColumnExpressionArgument:
        return self.resolve_property(prop).not_in(value)

    def apply_like(self, prop: str, value: str) -> ColumnExpressionArgument:
        sql_pattern = value.replace("*", "%")
        return self.resolve_property(prop).like(sql_pattern)

    def apply_ilike(self, prop: str, value: str) -> ColumnExpressionArgument:
        sql_pattern = value.replace("*", "%")
        return self.resolve_property(prop).ilike(sql_pattern)

    def resolve_property(self, prop_path: str) -> UnaryExpression:
        prop_path = self.resolve_alias(prop_path)
        model = self.model_class
        parts = prop_path.split(".")

        if len(parts) > 1:
            current = model

            for part in parts[:-1]:
                relationship = getattr(current, part)
                target_model = relationship.property.mapper.class_
                alias = aliased(target_model)

                if not relationship.property.uselist:
                    is_nullable = any(
                        column.nullable for column in relationship.property.local_columns
                    )
                    self.joins.append(
                        JoinExpression(target=alias, on=relationship, is_outer=is_nullable)
                    )
                current = alias

            return getattr(current, parts[-1])

        return getattr(model, prop_path)

    def apply_any(
        self, relationship_name, condition: ColumnElement | ColumnExpressionArgument
    ) -> Exists:
        relationship = getattr(self.model_class, relationship_name)
        related_model = relationship.property.mapper.class_
        alias = aliased(related_model)

        foreign_keys = list(relationship.property.local_remote_pairs)
        local_col, remote_col = foreign_keys[0]

        exists_clause = exists(alias).where(
            getattr(alias, remote_col.name) == getattr(self.model_class, local_col.name),
            self._adapt_condition(condition, alias),
        )
        return exists_clause

    def _adapt_condition(self, condition, alias):
        if isinstance(condition, BooleanClauseList):
            return condition.operator(
                *(self._adapt_condition(clause, alias) for clause in condition.clauses)
            )
        return condition.operator(getattr(alias, condition.left.key), condition.right.value)

    def apply_filter(self, query: Query, filter_expression: FilterExpression) -> Query:
        return query.filter(filter_expression.condition)

    def apply_joins(self, query: Query):
        for join_expr in self.joins:
            query = query.join(join_expr.target, join_expr.on, isouter=join_expr.is_outer)
        return query

    def apply_order_by(self, query: Query, order_by_expression: OrderByExpression) -> Query:
        fields = []
        for order_field in order_by_expression.fields:
            field = self.resolve_property(order_field.field_path)
            if order_field.direction == "-":
                field = field.desc()
            fields.append(field)
        return query.order_by(*fields)



import pytest
from django.db.models import Q

from requela.builders.django import DjangoQueryBuilder
from tests.django.models import User
from tests.django.utils import assert_statements_equal


@pytest.mark.parametrize(
    ("query_string", "expected"),
    [
        ("and(eq(name,John),eq(age,30))", User.objects.filter(Q(name="John") & Q(age=30))),
        (
            "and(eq(name,John),eq(age,30),eq(email,test@example.com))",
            User.objects.filter(Q(name="John") & Q(age=30) & Q(email="test@example.com")),
        ),
    ],
)
def test_logical_and(query_string, expected):
    builder = DjangoQueryBuilder(User)
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


@pytest.mark.parametrize(
    ("query_string", "expected"),
    [
        (
            "or(eq(name,John),eq(age,30))",
            User.objects.filter(Q(name="John") | Q(age=30)),
        ),
        (
            "or(eq(name,John),eq(age,30),eq(email,test@example.com))",
            User.objects.filter(Q(name="John") | Q(age=30) | Q(email="test@example.com")),
        ),
    ],
)
def test_logical_or(query_string, expected):
    builder = DjangoQueryBuilder(User)
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


def test_logical_not():
    builder = DjangoQueryBuilder(User)
    stmt = builder.build_query("not(eq(name,John))")
    assert_statements_equal(stmt, User.objects.filter(~Q(name="John")))


def test_nested_logical():
    builder = DjangoQueryBuilder(User)
    stmt = builder.build_query("and(or(eq(name,John),eq(age,30)),eq(email,test@example.com))")
    assert_statements_equal(
        stmt,
        User.objects.filter((Q(name="John") | Q(age=30)) & Q(email="test@example.com")),
    )


def test_nested_logical_with_not():
    builder = DjangoQueryBuilder(User)
    stmt = builder.build_query(
        "and(or(eq(name,John),eq(age,30)),not(in(email,(test@example.com,test2@example.com))))"
    )
    assert_statements_equal(
        stmt,
        User.objects.filter(
            (Q(name="John") | Q(age=30)) & ~Q(email__in=["test@example.com", "test2@example.com"])
        ),
    )



from datetime import date, datetime

import pytest

from requela.builders.django import DjangoQueryBuilder
from tests.django.models import Account, User
from tests.django.utils import assert_statements_equal


# Equality tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "name", "Ratatouille 123", User.objects.filter(name="Ratatouille 123")),
        (User, "age", 25, User.objects.filter(age=25)),
        (Account, "balance", 25.13, Account.objects.filter(balance=25.13)),
        (Account, "balance", -71.14, Account.objects.filter(balance=-71.14)),
        (Account, "balance", 33, Account.objects.filter(balance=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.filter(birth_date=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.filter(created_at=datetime.fromisoformat("2025-01-01T12:10:00+00:00")),
        ),
        (User, "is_active", "true", User.objects.filter(is_active=True)),
        (User, "is_active", "false", User.objects.filter(is_active=False)),
        (User, "name", "empty()", User.objects.filter(name="")),
        (User, "name", "null()", User.objects.filter(name__isnull=True)),
    ],
)
def test_comparison_eq(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"eq({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Not equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "name", "Ratatouille", User.objects.exclude(name="Ratatouille")),
        (User, "age", 25, User.objects.exclude(age=25)),
        (Account, "balance", 25.13, Account.objects.exclude(balance=25.13)),
        (Account, "balance", -71.14, Account.objects.exclude(balance=-71.14)),
        (Account, "balance", 33, Account.objects.exclude(balance=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.exclude(birth_date=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.exclude(created_at=datetime.fromisoformat("2025-01-01T12:10:00+00:00")),
        ),
        (User, "is_active", "true", User.objects.exclude(is_active=True)),
        (User, "is_active", "false", User.objects.exclude(is_active=False)),
        (User, "name", "empty()", User.objects.exclude(name="")),
        (User, "name", "null()", User.objects.filter(name__isnull=False)),
    ],
)
def test_comparison_ne(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"ne({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Less than tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, User.objects.filter(age__lt=25)),
        (Account, "balance", 25.13, Account.objects.filter(balance__lt=25.13)),
        (Account, "balance", -71.14, Account.objects.filter(balance__lt=-71.14)),
        (Account, "balance", 33, Account.objects.filter(balance__lt=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.filter(birth_date__lt=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.filter(
                created_at__lt=datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_lt(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"lt({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Greater than tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, User.objects.filter(age__gt=25)),
        (Account, "balance", 25.13, Account.objects.filter(balance__gt=25.13)),
        (Account, "balance", -71.14, Account.objects.filter(balance__gt=-71.14)),
        (Account, "balance", 33, Account.objects.filter(balance__gt=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.filter(birth_date__gt=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.filter(
                created_at__gt=datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_gt(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"gt({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Less than or equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, User.objects.filter(age__lte=25)),
        (Account, "balance", 25.13, Account.objects.filter(balance__lte=25.13)),
        (Account, "balance", -71.14, Account.objects.filter(balance__lte=-71.14)),
        (Account, "balance", 33, Account.objects.filter(balance__lte=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.filter(birth_date__lte=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.filter(
                created_at__lte=datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_lte(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"lte({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Greater than or equal tests
@pytest.mark.parametrize(
    ("model", "field", "value", "expected"),
    [
        (User, "age", 25, User.objects.filter(age__gte=25)),
        (Account, "balance", 25.13, Account.objects.filter(balance__gte=25.13)),
        (Account, "balance", -71.14, Account.objects.filter(balance__gte=-71.14)),
        (Account, "balance", 33, Account.objects.filter(balance__gte=33)),
        (
            User,
            "birth_date",
            "1974-06-29",
            User.objects.filter(birth_date__gte=date(1974, 6, 29)),
        ),
        (
            Account,
            "created_at",
            "2025-01-01T12:10:00+00:00",
            Account.objects.filter(
                created_at__gte=datetime.fromisoformat("2025-01-01T12:10:00+00:00")
            ),
        ),
    ],
)
def test_comparison_gte(model, field, value, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"gte({field},{value})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# In tests
@pytest.mark.parametrize(
    ("model", "field", "values", "expected"),
    [
        (User, "age", "(25,30,35)", User.objects.filter(age__in=[25, 30, 35])),
        (
            Account,
            "balance",
            "(25.13,50.00,75.25)",
            Account.objects.filter(balance__in=[25.13, 50.00, 75.25]),
        ),
        (
            User,
            "name",
            "(Alice,Bob,Charlie)",
            User.objects.filter(name__in=["Alice", "Bob", "Charlie"]),
        ),
    ],
)
def test_comparison_in(model, field, values, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"in({field},{values})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Out tests
@pytest.mark.parametrize(
    ("model", "field", "values", "expected"),
    [
        (User, "age", "(25,30,35)", User.objects.exclude(age__in=[25, 30, 35])),
        (
            Account,
            "balance",
            "(25.13,50.00,75.25)",
            Account.objects.exclude(balance__in=[25.13, 50.00, 75.25]),
        ),
        (
            User,
            "name",
            "(Alice,Bob,Charlie)",
            User.objects.exclude(name__in=["Alice", "Bob", "Charlie"]),
        ),
    ],
)
def test_comparison_out(model, field, values, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"out({field},{values})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# Like tests
@pytest.mark.parametrize(
    ("model", "field", "pattern", "expected"),
    [
        (User, "name", "Rat*", User.objects.filter(name__startswith="Rat")),
        (User, "name", "*ouille", User.objects.filter(name__endswith="ouille")),
        (User, "name", "*atato*", User.objects.filter(name__contains="atato")),
        (User, "name", "atato", User.objects.filter(name__contains="atato")),
    ],
)
def test_comparison_like(model, field, pattern, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"like({field},{pattern})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


# ILike tests
@pytest.mark.parametrize(
    ("model", "field", "pattern", "expected"),
    [
        (User, "name", "rat*", User.objects.filter(name__startswith="rat")),
        (User, "name", "*OUILLE", User.objects.filter(name__endswith="OUILLE")),
        (User, "name", "*AtAtO*", User.objects.filter(name__contains="AtAtO")),
        (User, "name", "AtAtO", User.objects.filter(name__contains="AtAtO")),
    ],
)
def test_comparison_ilike(model, field, pattern, expected):
    builder = DjangoQueryBuilder(model)
    query_string = f"ilike({field},{pattern})"
    stmt = builder.build_query(query_string)
    assert_statements_equal(stmt, expected)


def test_with_initial_query():
    builder = DjangoQueryBuilder(User)
    initial = User.objects.filter(age__gt=25)
    query_string = "eq(name,Ratatouille)"
    stmt = builder.build_query(query_string, initial)
    assert_statements_equal(stmt, User.objects.filter(age__gt=25, name="Ratatouille"))
